/**************************************************************************
       Title:   NTP Dual Clock
      Author:   Bruce E. Hall, w8bh.net
        Date:   02 Dec 2020
    Hardware:   Adafruit ESP32 Feather, ILI9341 TFT display module
    Software:   Arduino IDE 1.8.13 with Expressif ESP32 package
                TFT_eSPI Library
                ezTime Library
       Legal:   Copyright (c) 2020  Bruce E. Hall.
                Open Source under the terms of the MIT License.

  Description:   Dual UTC/Local NTP Clock with TFT display
                Time is refreshed via NTP every 30 minutes
                Optional time output to serial port
                Status indicator for time freshness & WiFi strength

                Before using, please update WIFI_SSID and WIFI_PWD
                with your personal WiFi credentials.  Also, modify
                TZ_RULE with your own Posix timezone string.

                see w8bh.net for a detailled, step-by-step tutorial

  Version Hx:   11/27/20  Initial GitHub commit
                11/28/20  added code to handle dropped WiFi connection
                11/30/20  showTimeDate() mod by John Price (WA2FZW)
                12/01/20  showAMPM() added by John Price (WA2FZW)

                07/02/2021 modified for mcufriend display by Jean-Marie  ON7EN

 **************************************************************************/


//#include <TFT_eSPI.h>                              // https://github.com/Bodmer/TFT_eSPI
#include <ezTime.h>                                // https://github.com/ropg/ezTime
#include <WiFi.h>

#define TITLE              "NTP TIME"
#define WIFI_SSID    "yourSSID"               
#define WIFI_PWD      "yourPassword"      
#define NTP_SERVER         "pool.ntp.org"          // time.nist.gov, pool.ntp.org, etc    
//#define TZ_RULE  "MSK-3MSD,M3.5.0/2,M10.5.0/3"
#define TZ_RULE            "CET-1CEST,M3.4.0/02:00:00,M10.5.0/03:00:00"//"EST5EDT,M3.2.0/2:00:00,M11.1.0/2:00:00"
uint16_t ID;
#define DEBUGLEVEL            INFO                 // NONE, ERROR, INFO, or DEBUG
#define PRINTED_TIME             1                 // 0=NONE, 1=UTC, or 2=LOCAL
#define TIME_FORMAT         COOKIE                 // COOKIE, ISO8601, RFC822, RFC850, RFC3339, RSS
#define BAUDRATE            115200                 // serial output baudrate
#define LEADING_ZERO         true                // show "01:00" vs " 1:00"
#define SYNC_MARGINAL         3600                 // orange status if no sync for 1 hour
#define SYNC_LOST            86400                 // red status if no sync for 1 day   
#define LOCAL_FORMAT_12HR    false                 // local time format 12hr "11:34" vs 24hr "23:34"
#define UTC_FORMAT_12HR      false                 // UTC time format 12 hr "11:34" vs 24hr "23:34"
#define DISPLAY_AMPM          false                // if true, show 'A' for AM, 'P' for PM
#define SCREEN_ORIENTATION       1                 // screen portrait mode:  use 1 or 3
#define TIMECOLOR         TFT_CYAN                 // color of 7-segment time display
#define DATECOLOR       TFT_YELLOW                 // color of displayed month & day
#define LABEL_FGCOLOR   TFT_YELLOW                 // color of label text
#define LABEL_BGCOLOR     TFT_BLUE                 // color of label background

#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

// ============ GLOBAL VARIABLES =====================================================
#include <MCUFRIEND_kbv.h>
#include<utility/mcufriend_shield.h>
MCUFRIEND_kbv tft;
//TFT_eSPI tft = TFT_eSPI();                         // display object
Timezone local;                                    // local timezone variable
time_t t, oldT;                                    // current & displayed UTC
time_t lt, oldLt;                                  // current & displayed local time
bool useLocalTime  = false;                        // temp flag used for display updates


// ============ DISPLAY ROUTINES =====================================================

void showClockStatus() {
  const int x = 290, y = 1, w = 28, h = 29, f = 2; // screen position & size
  int color;
  if (second() % 10) return;                       // update every 10 seconds
  int syncAge = now() - lastNtpUpdateTime();       // how long has it been since last sync?
  if (syncAge < SYNC_MARGINAL)                     // GREEN: time is good & in sync
    color = TFT_GREEN;
  else if (syncAge < SYNC_LOST)                    // ORANGE: sync is 1-24 hours old
    color = TFT_ORANGE;
  else color = TFT_RED;                            // RED: time is stale, over 24 hrs old
  if (WiFi.status() != WL_CONNECTED) {             //
    color = TFT_DARKGREY;                          // GRAY: WiFi connection was lost
    WiFi.disconnect();                             // so drop current connection
    WiFi.begin(WIFI_SSID, WIFI_PWD);               // and attempt to reconnect
  }
  tft.fillRoundRect(x - 5, y, w, h, 10, color);      // show clock status as a color
  tft.setTextSize(2);
  tft.setTextColor(TFT_BLACK);
  tft.setCursor(x-2, y + 6);
  tft.print(-WiFi.RSSI());          // WiFi strength as a positive value
}

/*
    Modified by John Price (WA2FZW)

      In the original code, this was an empty function. I added code to display either
      an 'A' or 'P' to the right of the local time

*/

void showAMPM ( int hr, int x, int y )
{
  char  ampm;                                      // Will be either 'A' or 'P'
  if ( hr <= 11 )                                  // If the hour is 11 or less
    ampm = 'A';                                    // It's morning
  else                                             // Otherwise,
    ampm = 'P';
  tft.setCursor(x, y);// It's afternoon
  tft.print ( ampm);                  // Show AM/PM indicator
}

void showTime(time_t t, bool hr12, int x, int y) {
  tft.setTextSize(5);
  const int f = 7;                                 // screen font
  tft.setTextColor(TFT_CYAN);          // set time color
  int h = hour(t); int m = minute(t); int s = second(t); // get hours, minutes, and seconds
  if (hr12) {                                      // if using 12hr time format,
    if (DISPLAY_AMPM) showAMPM(h, x + 220, y + 14); // show AM/PM indicator
    if (h == 0) h = 12;                            // 00:00 becomes 12:00
    if (h > 12) h -= 12;                           // 13:00 becomes 01:00
  }
  if (h < 10) {                                    // is hour a single digit?
    if ((!hr12) || (LEADING_ZERO))
    { tft.setCursor(x, y);// 24hr format: always use leading 0
      x = tft.print("0");
    }            // show leading zero for hours
    else {
      tft.setTextColor(TFT_BLACK);       // black on black text
      tft.setCursor(x, y);
      x += tft.print("8");                // will erase the old digit
      tft.setTextColor(TFT_BLACK);
    }
  }
  tft.fillRect(x, y, 220, 50, TFT_BLACK);  x = x + 10;
  tft.setCursor(x, y + 10);
  if (h < 10) x = tft.print("0");
  x = tft.print(h);                    // hours
  tft.setCursor(x + 60, y + 10);
  //  x= tft.print(":");                     // show ":"
  tft.setCursor(x + 88, y + 10);
  if (m < 10) x = tft.print("0");        // leading zero for minutes
  // tft.setCursor(x+80, y+10);
  x = tft.print(m);                    // show minutes
  tft.setCursor(x + 135, y + 10);
  // x= tft.print(":");                     // show ":"
  // if (s<10) x+= tft.print("0");           // add leading zero for seconds
  tft.setCursor(x + 155, y + 10);
  if (s < 10) x += tft.print("0");
  x += tft.print(s);                    // show seconds
}

void showDate(time_t t, int x, int y) {
  const int f = 4, yspacing = 30;                  // screen font, spacing
  const char* months[] = {"JAN", "FEB", "MAR",
                          "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT",
                          "NOV", "DEC"
                         };
  tft.setTextColor(DATECOLOR, TFT_BLACK);
  tft.setTextSize(3);
  int m = month(t), d = day(t);                    // get date components
  tft.fillRect(x, y, 50, 60, TFT_BLACK);
  tft.setCursor(x, y); tft.print(d);      // erase previous date
  // show month on top
  y += yspacing;                                   // put day below month
  //if (d<10) x+=tft.drawNumber(0,x,y,f);            // draw leading zero for day
  tft.setCursor(x, y); tft.print(months[m - 1]);               // draw day
}

void showTimeZone (int x, int y) {
  const int f = 4;
  tft.setTextSize(2);// text font
  tft.setTextColor(LABEL_FGCOLOR, LABEL_BGCOLOR);  // set text colors
  tft.fillRect(x, y, 80, 28, LABEL_BGCOLOR);       // erase previous TZ
  if (!useLocalTime)
  {
    tft.setCursor(x, y);    // UTC time
    tft.print("UTC");
  }
  else
  { tft.setCursor(x, y); tft.print(local.getTimezoneName()); // show local time zone
  }
}
void showTimeDate(time_t t, time_t oldT, bool hr12, int x, int y) {
  showTime(t, hr12, x, y);                         // display time HH:MM:SS
  if ((!oldT) || (hour(t) != hour(oldT)))          // did hour change?
    showTimeZone(x, y - 42);                       // update time zone
  if ((!oldT) || (day(t) != day(oldT)))            // did date change? (Thanks John WA2FZW!)
    showDate(t, x + 250, y);                       // update date
}

void updateDisplay() {
  t = now();                                       // check latest time
  if (t != oldT) {                                 // are we in a new second yet?
    lt = local.now();                              // keep local time current
    useLocalTime = true;                           // use local timezone
    showTimeDate(lt, oldLt, LOCAL_FORMAT_12HR, 10, 46); // show new local time
    useLocalTime = false;                          // use UTC timezone
    showTimeDate(t, oldT, UTC_FORMAT_12HR, 10, 172); // show new UTC time
    showClockStatus();                             // and clock status
    printTime();                                   // send timestamp to serial port
    oldT = t; oldLt = lt;                          // remember currently displayed time
  }
}


void newDualScreen() {
  tft.fillScreen(TFT_BLACK);                       // start with empty screen
  tft.fillRoundRect(0, 0, 319, 32, 10, LABEL_BGCOLOR); // title bar for local time
  tft.fillRoundRect(0, 126, 319, 32, 10, LABEL_BGCOLOR); // title bar for UTC
  tft.setTextColor(LABEL_FGCOLOR, LABEL_BGCOLOR);  // set label colors
  tft.setCursor(100, 8); tft.print(TITLE);             // show title at top
  tft.drawRoundRect(0, 0, 319, 110, 10, TFT_WHITE); // draw edge around local time
  tft.drawRoundRect(0, 126, 319, 110, 10, TFT_WHITE); // draw edge around UTC
}

void startupScreen() {
  tft.fillScreen(TFT_BLACK);                       // start with empty screen
  tft.fillRoundRect(0, 0, 319, 30, 10, LABEL_BGCOLOR); // title bar
  tft.drawRoundRect(0, 0, 319, 239, 10, TFT_WHITE); // draw edge screen
  tft.setTextColor(LABEL_FGCOLOR, LABEL_BGCOLOR);  // set label colors
  //tft.drawCentreString(TITLE,160,2,4);             // show sketch title on screen
  tft.setCursor(160, 4); tft.print(TITLE);
  tft.setTextColor(LABEL_FGCOLOR, TFT_BLACK);      // set text color
}


// ============ MISC ROUTINES =====================================================

void showConnectionProgress() {
  int elapsed = 0;
  tft.setTextSize(2);
  tft.setCursor(5, 50); tft.print("WiFi starting");
  while (WiFi.status() != WL_CONNECTED) {          // while waiting for connection
    tft.setCursor(30, 50); tft.print(elapsed++);          // show we are trying!
    delay(1000);
  }
  tft.setCursor(5, 100); tft.print("IP = " +                        // connected to LAN now
                                   WiFi.localIP().toString());             // so show IP address
  elapsed = 0;
  tft.setCursor(5, 140); tft.print("Waiting for NTP");      // Now get NTP info
  while (timeStatus() != timeSet) {                // wait until time retrieved
    events();                                      // allow ezTime to work
    tft.setCursor(230, 140); tft.print(elapsed++);         // show we are trying
    delay(1000);
  }
}

void printTime() {                                 // print time to serial port
  if (!PRINTED_TIME) return;                       // option 0: dont print
  Serial.print("TIME: ");
  if (PRINTED_TIME == 1)
    Serial.println(dateTime(TIME_FORMAT));         // option 1: print UTC time
  else
    Serial.println(local.dateTime(TIME_FORMAT));   // option 2: print local time
}


// ============ MAIN PROGRAM ===================================================

void setup() {
  //tft.init();   // initialize LCD screen object
  tft.reset();
  ID = tft.readID();
  tft.begin(ID);
  tft.setRotation(SCREEN_ORIENTATION);             // landscape screen orientation
  startupScreen();                                 // show title
  Serial.begin(BAUDRATE);                          // open serial port
  setDebug(DEBUGLEVEL);                            // enable NTP debug info
  setServer(NTP_SERVER);                           // set NTP server
  WiFi.begin(WIFI_SSID, WIFI_PWD);                 // start WiFi
  showConnectionProgress();                        // WiFi and NTP may take time
  local.setPosix(TZ_RULE);                         // estab. local TZ by rule
  newDualScreen();                                 // show title & labels
}

void loop() {
  events();                                        // get periodic NTP updates
  updateDisplay();                                 // update clock every second
}
